% This script solves three optimal control problems for the "car on a hill"
% system. The third optimal control problem uses a terminal cost function
% that is a blend of the terminal costs of the first two. Then, it uses
% linear Bellman combination to solve the third optimal control problem
% instantly and compares the results. It also compares to a naive blend
% of the policies.
%
% Note that the three optimal control problems are solved using Emo
% Todorov's z-iteration algorithm. 
%
clear all;

mdp_params.dt = 0.05;          % timestep
mdp_params.sigma = 1;          % velocity noise: stdev per unit time
mdp_params.sigma_actual = 0;

mdp_params.nstdB = 3;          % number of standard deviations, new algorithm (B)

mdp_params.maxx = 3;           % extent of x-grid
mdp_params.maxv = 9;           % extent of v-grid

mdp_params.nx = 301;           % number of x-grid points
mdp_params.nv = 201;           % number of v-grid points

sclq = 1;          % scale of state costs
sclqend = 4;    % scale of terminal costs
epsx = 0.04;    % end-position tolerance
epsv = 8;         % end-velocity tolerance

fuzzy_target = 0;

sclp = 2;           % scale of potential function
mdp_params.grav = 9.8;         % gravitational acceleration
mdp_params.damp = 0.5;         % damping coefficient

% functions used to compute the dynamics of the "car"
p0 = @(x)(sclp*(1-exp(-0.5*x.^2)));             % potential function
p1 = @(x)(sclp*x.*exp(-0.5*x.^2));              % slope(x) = tan(al(x))
cos_al = @(x)(1 ./ sqrt(1+p1(x).^2));           % cos(al(x))
sin_al = @(x)(sign(x) ./ sqrt(1+1./p1(x).^2));  % sin(al(x))

mdp_params.NT = 300;

% here the weights are set by hand
wts = [ 0.5 0.5 ];

% setup grids and parameters for Z-iteration
gx = linspace(-mdp_params.maxx, mdp_params.maxx, mdp_params.nx);
gv = linspace(-mdp_params.maxv, mdp_params.maxv, mdp_params.nv);
dx = gx(2) - gx(1);
dv = gv(2) - gv(1);

nsB = 2*round(mdp_params.nstdB*sqrt(mdp_params.dt)*mdp_params.sigma/dv) + 1;

[X, V] = ndgrid(gx, gv);

% compute costs
Q = mdp_params.dt*ones(mdp_params.nx, mdp_params.nv)*sclq;
itrg = ( (abs(X-fuzzy_target)<=epsx) & (abs(V)<=epsv) );
Q(itrg) = 0;

% compute continuous passive dynamics in grid units
nxtx = (X + mdp_params.dt * V.*cos_al(X)) / dx;
nxtv = (V + mdp_params.dt * (-mdp_params.grav*sin_al(X) -mdp_params.damp*V)) / dv;
nxtx = min(mdp_params.nx, max(1, nxtx + (mdp_params.nx-1)/2+1));
nxtv = min(mdp_params.nv, max(1, nxtv + (mdp_params.nv-1)/2+1));
nxtvr = round(nxtv);

% find nearest discrete positions, compute probabilities
NX{1} = ones(mdp_params.nx,mdp_params.nv);
NX{2} = ones(mdp_params.nx,mdp_params.nv);
PX = ones(mdp_params.nx,mdp_params.nv) * 0.5;
ieq = ( abs(round(nxtx)-nxtx)<eps );
NX{1}(ieq) = round(nxtx(ieq));
NX{2}(ieq) = round(nxtx(ieq));
NX{1}(~ieq) = floor(nxtx(~ieq));
NX{2}(~ieq) = ceil(nxtx(~ieq));
PX(~ieq) = NX{2}(~ieq) - nxtx(~ieq);

% find nearest discrete velocities, compute probabilities, new method (B)
NVB = cell(nsB,1);
PVB = cell(nsB,1);
probtot = zeros(mdp_params.nx,mdp_params.nv);
for is=1:nsB
    NVB{is} = nxtvr + is-(nsB+1)/2;
    PVB{is} = exp( -(NVB{is}-nxtv).^2/(2*mdp_params.dt*mdp_params.sigma^2/dv^2) );
    probtot = probtot + PVB{is};
    NVB{is} = min(mdp_params.nv, max(1, NVB{is}));
end
for is=1:nsB
    PVB{is} = PVB{is} ./ probtot;
end

% --------------------- first boundary condition  ----------------------
%

Qend1 = sclqend*((X).^2+ (V-1).^2);
[Z1,UUdif1] = mdp_zloop(mdp_params,Q,Qend1,nsB,NVB,PVB,NX,dv,PX);

% --------------------- second boundary condition  ---------------------
%

Qend2 = sclqend*(X.^2+ (V-2).^2);
[Z2,UUdif2] = mdp_zloop(mdp_params,Q,Qend2,nsB,NVB,PVB,NX,dv,PX);


% ---------------------- blend boundary condition  ---------------------
%

QendBlend = -log(wts(1)*exp(-Qend1)+wts(2)*exp(-Qend2));

[mQB,mI1] = min(QendBlend);
[mQB2,mI2] = min(mQB);

disp('min at');
dx*mI1(mI2) - mdp_params.maxx
dv*mI2 - mdp_params.maxv

[Zopt,UUdifopt] = mdp_zloop(mdp_params,Q,QendBlend,nsB,NVB,PVB,NX,...
    dv,PX);

UUdifnaive = cell(1,mdp_params.NT-1);
for i = 1:mdp_params.NT-1
    UUdifnaive{i} = wts(1)*UUdif1{i} + wts(2)*UUdif2{i};
end

UUdifzblend = cell(1,mdp_params.NT-1);
for i = 1:mdp_params.NT-1
    z = wts(1)*Z1{i} + wts(2)*Z2{i};
    a1 = wts(1)*Z1{i}./z;
    a2 = wts(2)*Z2{i}./z;
    UUdifzblend{i} = a1.*UUdif1{i} + a2.*UUdif2{i};
end

display('flip thru the time varying policy')
display('for the composite terminal cost')
display('one is computed using z-iteration')
display('one is computed using linear Bellman combination')
display('and the third using a naive blend')

figure(444);
for i = 1:mdp_params.NT-1
    subplot(1,3,1);
    imagesc(UUdifopt{i});
    title(sprintf('Z-iteration result (ground truth), t = %d',i));
   
    m1 = max(max(abs(UUdifzblend{i}-UUdifopt{i})));
    m2 = max(max(abs(UUdifnaive{i}-UUdifopt{i})));
    m3 = max(m1,m2);
    
    subplot(1,3,2);
    imagesc(abs(UUdifzblend{i}-UUdifopt{i}),[0,m3]);
    title('Linear Bellman Combination (difference from ground truth)');
    

    subplot(1,3,3);
    imagesc(abs(UUdifnaive{i}-UUdifopt{i}),[0,m3]);
    title('Naive Combination (difference from ground truth)');
    
    pause
end

